@@ -181,6 +181,7 @@ struct cbinfo {
 };
 
 /* s3backer_store functions */
+static int block_cache_create_threads(struct s3backer_store *s3b);
 static int block_cache_meta_data(struct s3backer_store *s3b, off_t *file_sizep, u_int *block_sizep);
 static int block_cache_set_mount_token(struct s3backer_store *s3b, int32_t *old_valuep, int32_t new_value);
 static int block_cache_read_block(struct s3backer_store *s3b, s3b_block_t block_num, void *dest,
@@ -234,7 +235,6 @@ block_cache_create(struct block_cache_conf *config, struct s3backer_store *inner
     struct s3backer_store *s3b;
     struct block_cache_private *priv;
     struct cache_entry *entry;
-    pthread_t thread;
     int r;
 
     /* Initialize s3backer_store structure */
@@ -243,6 +243,7 @@ block_cache_create(struct block_cache_conf *config, struct s3backer_store *inner
         (*config->log)(LOG_ERR, "calloc(): %s", strerror(r));
         goto fail0;
     }
+    s3b->create_threads = block_cache_create_threads;
     s3b->meta_data = block_cache_meta_data;
     s3b->set_mount_token = block_cache_set_mount_token;
     s3b->read_block = block_cache_read_block;
@@ -301,22 +302,11 @@ block_cache_create(struct block_cache_conf *config, struct s3backer_store *inner
     pthread_mutex_lock(&priv->mutex);
     S3BCACHE_CHECK_INVARIANTS(priv);
 
-    /* Create threads */
-    for (priv->num_threads = 0; priv->num_threads < config->num_threads; priv->num_threads++) {
-        if ((r = pthread_create(&thread, NULL, block_cache_worker_main, priv)) != 0)
-            goto fail10;
-    }
-
     /* Done */
     pthread_mutex_unlock(&priv->mutex);
     return s3b;
 
-fail10:
-    priv->stopping = 1;
-    while (priv->num_threads > 0) {
-        pthread_cond_broadcast(&priv->worker_work);
-        pthread_cond_wait(&priv->worker_exit, &priv->mutex);
-    }
+fail9:
     if (config->cache_file != NULL) {
         while ((entry = TAILQ_FIRST(&priv->cleans)) != NULL) {
             TAILQ_REMOVE(&priv->cleans, entry, link);
@@ -324,7 +314,6 @@ fail10:
         }
         s3b_dcache_close(priv->dcache);
     }
-fail9:
     s3b_hash_destroy(priv->hashtable);
 fail8:
     pthread_cond_destroy(&priv->write_complete);
@@ -401,6 +390,34 @@ block_cache_dcache_load(void *arg, s3b_block_t dslot, s3b_block_t block_num, con
     return 0;
 }
 
+static int
+block_cache_create_threads(struct s3backer_store *s3b)
+{
+    struct block_cache_private *const priv = s3b->data;
+    struct block_cache_conf *const config = priv->config;
+    pthread_t thread;
+    int r;
+
+    /* Create threads in lower layer */
+    if ((r = (*priv->inner->create_threads)(priv->inner)) != 0)
+        return r;
+
+    /* Grab lock */
+    pthread_mutex_lock(&priv->mutex);
+    S3BCACHE_CHECK_INVARIANTS(priv);
+
+    /* Create threads */
+    while (priv->num_threads < config->num_threads) {
+        if ((r = pthread_create(&thread, NULL, block_cache_worker_main, priv)) != 0)
+            goto fail;
+        priv->num_threads++;
+    }
+
+fail:
+    pthread_mutex_unlock(&priv->mutex);
+    return r;
+}
+
 static int
 block_cache_meta_data(struct s3backer_store *s3b, off_t *file_sizep, u_int *block_sizep)
 {
@@ -495,6 +512,16 @@ block_cache_get_stats(struct s3backer_store *s3b, struct block_cache_stats *stat
     pthread_mutex_unlock(&priv->mutex);
 }
 
+void
+block_cache_clear_stats(struct s3backer_store *s3b)
+{
+    struct block_cache_private *const priv = s3b->data;
+
+    pthread_mutex_lock(&priv->mutex);
+    memset(&priv->stats, 0, sizeof(priv->stats));
+    pthread_mutex_unlock(&priv->mutex);
+}
+
 static int
 block_cache_list_blocks(struct s3backer_store *s3b, block_list_func_t *callback, void *arg)
 {
@@ -545,6 +572,12 @@ block_cache_read(struct block_cache_private *const priv, s3b_block_t block_num,
     pthread_mutex_lock(&priv->mutex);
     S3BCACHE_CHECK_INVARIANTS(priv);
 
+    /* Sanity check */
+    if (priv->num_threads == 0) {
+        (*config->log)(LOG_ERR, "block_cache_read(): no threads created yet");
+        return ENOTCONN;
+    }
+
     /* Update count of block(s) read sequentially by the upper layer */
     if (block_num == priv->seq_last + 1) {
         priv->seq_count++;
@@ -773,6 +806,7 @@ block_cache_write(struct block_cache_private *const priv, s3b_block_t block_num,
 {
     struct block_cache_conf *const config = priv->config;
     struct cache_entry *entry;
+    int partial_miss = 0;
     int r;
 
     /* Sanity check */
@@ -786,6 +820,11 @@ block_cache_write(struct block_cache_private *const priv, s3b_block_t block_num,
 again:
     /* Sanity check */
     S3BCACHE_CHECK_INVARIANTS(priv);
+    if (priv->num_threads == 0) {
+        (*config->log)(LOG_ERR, "block_cache_write(): no threads created yet");
+        r = ENOTCONN;
+        goto fail;
+    }
 
     /* Find cache entry */
     if ((entry = s3b_hash_get(priv->hashtable, block_num)) != NULL) {
@@ -827,7 +866,8 @@ again:
             if ((r = block_cache_write_data(priv, entry, src, off, len)) != 0)
                 (*config->log)(LOG_ERR, "error updating dirty block! %s", strerror(r));
             entry->dirty = 1;
-            priv->stats.write_hits++;
+            if (!partial_miss)
+                priv->stats.write_hits++;
             break;
         default:
             assert(0);
@@ -843,6 +883,8 @@ again:
     if (off != 0 || len != config->block_size) {
         if ((r = block_cache_do_read(priv, block_num, 0, 0, NULL, 0)) != 0)
             goto fail;
+        if (partial_miss++ == 0)
+            priv->stats.write_misses++;
         goto again;
     }
 
@@ -1279,9 +1321,10 @@ block_cache_verified(struct block_cache_private *priv, struct cache_entry *entry
     assert(entry->verify);
     assert(ENTRY_GET_STATE(entry) == CLEAN2 || ENTRY_GET_STATE(entry) == READING2);
 
-    /* Give back some memory; if we can't no big deal */
-    if ((new_entry = realloc(entry, sizeof(*entry))) == NULL)
+    /* Allocate new, smaller entry; if we can't no big deal */
+    if ((new_entry = malloc(sizeof(*entry))) == NULL)
         goto done;
+    memcpy(new_entry, entry, sizeof(*entry));
 
     /* Update all references that point to the entry */
     s3b_hash_put(priv->hashtable, new_entry);
@@ -1289,6 +1332,7 @@ block_cache_verified(struct block_cache_private *priv, struct cache_entry *entry
         TAILQ_REMOVE(&priv->cleans, entry, link);
         TAILQ_INSERT_TAIL(&priv->cleans, new_entry, link);
     }
+    free(entry);
     entry = new_entry;
 
 done:
