@@ -129,6 +129,7 @@ struct cbinfo {
 };
 
 /* s3backer_store functions */
+static int ec_protect_create_threads(struct s3backer_store *s3b);
 static int ec_protect_meta_data(struct s3backer_store *s3b, off_t *file_sizep, u_int *block_sizep);
 static int ec_protect_set_mount_token(struct s3backer_store *s3b, int32_t *old_valuep, int32_t new_value);
 static int ec_protect_read_block(struct s3backer_store *s3b, s3b_block_t block_num, void *dest,
@@ -182,6 +183,7 @@ ec_protect_create(struct ec_protect_conf *config, struct s3backer_store *inner)
         (*config->log)(LOG_ERR, "calloc(): %s", strerror(r));
         goto fail0;
     }
+    s3b->create_threads = ec_protect_create_threads;
     s3b->meta_data = ec_protect_meta_data;
     s3b->set_mount_token = ec_protect_set_mount_token;
     s3b->read_block = ec_protect_read_block;
@@ -234,6 +236,14 @@ fail0:
     return NULL;
 }
 
+static int
+ec_protect_create_threads(struct s3backer_store *s3b)
+{
+    struct ec_protect_private *const priv = s3b->data;
+
+    return (*priv->inner->create_threads)(priv->inner);
+}
+
 static int
 ec_protect_meta_data(struct s3backer_store *s3b, off_t *file_sizep, u_int *block_sizep)
 {
@@ -306,6 +316,16 @@ ec_protect_get_stats(struct s3backer_store *s3b, struct ec_protect_stats *stats)
     pthread_mutex_unlock(&priv->mutex);
 }
 
+void
+ec_protect_clear_stats(struct s3backer_store *s3b)
+{
+    struct ec_protect_private *const priv = s3b->data;
+
+    pthread_mutex_lock(&priv->mutex);
+    memset(&priv->stats, 0, sizeof(priv->stats));
+    pthread_mutex_unlock(&priv->mutex);
+}
+
 static int
 ec_protect_list_blocks(struct s3backer_store *s3b, block_list_func_t *callback, void *arg)
 {
