@@ -106,6 +106,7 @@ struct list_blocks {
  ****************************************************************************/
 
 static print_stats_t s3b_config_print_stats;
+static clear_stats_t s3b_config_clear_stats;
 
 static int parse_size_string(const char *s, uintmax_t *valp);
 static void unparse_size_string(char *buf, size_t bmax, uintmax_t value);
@@ -151,6 +152,7 @@ static struct s3b_config config = {
         .region=                NULL,
         .bucket=                NULL,
         .sse=                   NULL,
+        .blockHashPrefix=       0,
         .prefix=                S3BACKER_DEFAULT_PREFIX,
         .accessType=            S3BACKER_DEFAULT_ACCESS_TYPE,
         .authVersion=           S3BACKER_DEFAULT_AUTH_VERSION,
@@ -371,6 +373,11 @@ static const struct fuse_opt option_list[] = {
         .templ=     "--minWriteDelay=%u",
         .offset=    offsetof(struct s3b_config, ec_protect.min_write_delay),
     },
+    {
+        .templ=     "--blockHashPrefix",
+        .offset=    offsetof(struct s3b_config, http_io.blockHashPrefix),
+        .value=     1
+    },
     {
         .templ=     "--prefix=%s",
         .offset=    offsetof(struct s3b_config, http_io.prefix),
@@ -582,6 +589,7 @@ s3backer_get_config(int argc, char **argv)
 
     /* Set up fuse_ops callbacks */
     config.fuse_ops.print_stats = s3b_config_print_stats;
+    config.fuse_ops.clear_stats = s3b_config_clear_stats;
     config.fuse_ops.s3bconf = &config;
 
     /* Debug */
@@ -593,7 +601,7 @@ s3backer_get_config(int argc, char **argv)
 }
 
 /*
- * Create the s3backer_store used at runtime. This method is invoked by fuse_op_init().
+ * Create the s3backer_store used at runtime.
  */
 struct s3backer_store *
 s3backer_create_store(struct s3b_config *conf)
@@ -778,6 +786,22 @@ s3b_config_print_stats(void *prarg, printer_t *printer)
     (*printer)(prarg, "%-28s %u\n", "out_of_memory_errors", total_oom);
 }
 
+static void
+s3b_config_clear_stats(void)
+{
+    /* Clear HTTP stats */
+    if (http_io_store != NULL)
+        http_io_clear_stats(http_io_store);
+
+    /* Clear EC protection stats */
+    if (ec_protect_store != NULL)
+        ec_protect_clear_stats(ec_protect_store);
+
+    /* Clear block cache stats */
+    if (block_cache_store != NULL)
+        block_cache_clear_stats(block_cache_store);
+}
+
 static int
 parse_size_string(const char *s, uintmax_t *valp)
 {
@@ -959,7 +983,8 @@ validate_config(void)
         config.http_io.accessId = NULL;
 
     /* If no accessId, only read operations will succeed */
-    if (config.http_io.accessId == NULL && !config.fuse_ops.read_only && !customBaseURL && config.http_io.ec2iam_role == NULL) {
+    if (!config.test && config.http_io.accessId == NULL
+      && !config.fuse_ops.read_only && !customBaseURL && config.http_io.ec2iam_role == NULL) {
         warnx("warning: no `accessId' specified; only read operations will succeed");
         warnx("you can eliminate this warning by providing the `--readOnly' flag");
     }
@@ -1229,7 +1254,7 @@ validate_config(void)
     }
     if (config.file_size_str != NULL) {
         if (parse_size_string(config.file_size_str, &value) == -1 || value == 0) {
-            warnx("invalid file size `%s'", config.block_size_str);
+            warnx("invalid file size `%s'", config.file_size_str);
             return -1;
         }
         config.file_size = value;
@@ -1634,6 +1659,7 @@ dump_config(void)
     (*config.log)(LOG_DEBUG, "%24s: \"%s\"", "region", config.http_io.region);
     (*config.log)(LOG_DEBUG, "%24s: \"%s\"", config.test ? "testdir" : "bucket", config.http_io.bucket);
     (*config.log)(LOG_DEBUG, "%24s: \"%s\"", "prefix", config.http_io.prefix);
+    (*config.log)(LOG_DEBUG, "%24s: %s", "blockHashPrefix", config.http_io.blockHashPrefix ? "true" : "false");
     (*config.log)(LOG_DEBUG, "%24s: \"%s\"", "defaultContentEncoding",
       config.http_io.default_ce != NULL ? config.http_io.default_ce : "(none)");
     (*config.log)(LOG_DEBUG, "%24s: %s", "list_blocks", config.list_blocks ? "true" : "false");
@@ -1777,6 +1803,7 @@ usage(void)
     fprintf(stderr, "\t--%-27s %s\n", "blockCacheTimeout=MILLIS", "Block cache entry timeout (zero = infinite)");
     fprintf(stderr, "\t--%-27s %s\n", "blockCacheWriteDelay=MILLIS", "Block cache maximum write-back delay");
     fprintf(stderr, "\t--%-27s %s\n", "blockSize=SIZE", "Block size (with optional suffix 'K', 'M', 'G', etc.)");
+    fprintf(stderr, "\t--%-27s %s\n", "blockHashPrefix", "Prepend hash to block names for even distribution");
     fprintf(stderr, "\t--%-27s %s\n", "cacert=FILE", "Specify SSL certificate authority file");
     fprintf(stderr, "\t--%-27s %s\n", "compress[=LEVEL]", "Enable block compression, with 1=fast up to 9=small");
     fprintf(stderr, "\t--%-27s %s\n", "debug", "Enable logging of debug messages");
