@@ -119,6 +119,25 @@ typedef int         check_cancel_t(void *arg, s3b_block_t block_num);
 /* Backing store instance structure */
 struct s3backer_store {
 
+    /*
+     * Create any background pthreads that may be required.
+     *
+     * This must be invoked prior to any of the following functions:
+     *
+     *      o block_read
+     *      o block_read_part
+     *      o block_write
+     *      o block_write_part
+     *
+     * It should be invoked after the initial process fork() because it may create pthreads.
+     *
+     * Returns:
+     *
+     *  0       Success
+     *  Other   Other error
+     */
+    int         (*create_threads)(struct s3backer_store *s3b);
+
     /*
      * Get meta-data associated with the underlying store.
      *
@@ -166,6 +185,7 @@ struct s3backer_store {
      *      - Return EEXIST; the block may or may not also be read normally into *dest
      *
      * Returns zero on success or a (positive) errno value on error.
+     * May return ENOTCONN if create_threads() has not yet been invoked.
      */
     int         (*read_block)(struct s3backer_store *s3b, s3b_block_t block_num, void *dest,
                   u_char *actual_md5, const u_char *expect_md5, int strict);
@@ -174,6 +194,7 @@ struct s3backer_store {
      * Read part of one block.
      *
      * Returns zero on success or a (positive) errno value on error.
+     * May return ENOTCONN if create_threads() has not yet been invoked.
      */
     int         (*read_block_part)(struct s3backer_store *s3b, s3b_block_t block_num, u_int off, u_int len, void *dest);
 
@@ -189,6 +210,7 @@ struct s3backer_store {
      * parameter of read_block(); if the block is all zeroes, md5 will be zeroed.
      *
      * Returns zero on success or a (positive) errno value on error.
+     * May return ENOTCONN if create_threads() has not yet been invoked.
      */
     int         (*write_block)(struct s3backer_store *s3b, s3b_block_t block_num, const void *src, u_char *md5,
                   check_cancel_t *check_cancel, void *arg);
@@ -197,6 +219,7 @@ struct s3backer_store {
      * Write part of one block.
      *
      * Returns zero on success or a (positive) errno value on error.
+     * May return ENOTCONN if create_threads() has not yet been invoked.
      */
     int         (*write_block_part)(struct s3backer_store *s3b, s3b_block_t block_num, u_int off, u_int len, const void *src);
 
